// DirectX Boost related stuff.

if not WindowsTheComputer then return end
if srb2p then return end

local directxboost = CV_FindVar("DirectX")
local mobjscale = CV_FindVar("kartgrow")
local dbg = CV_FindVar("windbg")

local directplay = CV_RegisterVar({
	name = "dxcoop",
	defaultvalue = 0,
	flags = CV_NETVAR | CV_SHOWMODIF | CV_CALL,
	PossibleValue = {disabled=0, enabled=1},
	func = WindowsTheComputer.DoKartCallBackSound
})

local directplayrangeh = CV_RegisterVar({
	name = "dxcooprangeh",
	defaultvalue = 8,
	flags = CV_NETVAR | CV_SHOWMODIF | CV_CALL,
	PossibleValue = {MIN = 4, MAX = 32},
	func = WindowsTheComputer.DoKartCallBackSound
})

local directplayrangev = CV_RegisterVar({
	name = "dxcooprangev",
	defaultvalue = 3,
	flags = CV_NETVAR | CV_SHOWMODIF | CV_CALL,
	PossibleValue = {MIN = 1, MAX = 32},
	func = WindowsTheComputer.DoKartCallBackSound
})

local function IsSpecialHyper(p)
	local customhyper = WindowsTheComputer.HyperDef
	for i = 1, #customhyper do
		if customhyper[i](p) then return true else return false end
	end
	return false
end

COM_AddCommand("dxcoopeffect", function(player, arg)
	if not directxboost.value then
		CONS_Printf(player, "DirectX Boost is disabled. Please enable it to use this command.")
		return
	end
	if arg then
		arg = arg:lower()
		if arg == "1" or arg == "enable" or arg == "yes" then
			player.dxcooptoggle = true
			CONS_Printf(player, "You have opted in for effects from a nearby DirectX Boost.")
			S_StartSound(player.mo, sfx_addfil, player)
		elseif arg == "0" or arg == "disable" or arg == "no" then
			player.dxcooptoggle = false
			CONS_Printf(player, "You have opted out of effects from a nearby DirectX Boost.")
			S_StartSound(player.mo, sfx_notadd, player)
		else
			if player.dxcooptoggle then CONS_Printf(player, "You are opted in for effects from a nearby DirectX Boost.")
			elseif not player.dxcooptoggle then CONS_Printf(player, "You are opted out of effects from a nearby DirectX Boost.") end
		end
	else
		CONS_Printf(player, "When someone as Windows performs a DirectX Boost close enough to you, you also will be affected by it. Currently\x82\ " + player.dxcooptoggle + "\x80\.")
	end
end)

local function soundblaster(win)
	if win.powers[pw_super] != 0 or (win.solchar and win.solchar.istransformed and win.solchar.istransformed > 1) then S_StartSound(win.mo, sfx_pinbl2)
	elseif P_PlayerInPain(win) then S_StartSound(win.mo, sfx_hvrres)
	else S_StartSound(win.mo, sfx_hvrjmp) end
end

// P_RadiusAttack, but for everything damagable - monitors and alike.
// Simply miniature XPerience Blasts. Somehow went missing during 1.6dev.
local function DXAoE(win, range)
	for dll in mobjs.iterate()
		if R_PointToDist2(win.mo.x, win.mo.y, dll.x, dll.y) > range then continue
		elseif dll.type == MT_METALSONIC_BATTLE and dll.state == S_METALSONIC_FLOAT then
			dll.state = S_METALSONIC_PAIN
			if dll.health > 1 then dll.health = $ - 1
			elseif dll.health <= 1 then
				A_BossDeath(dll)
				dll.momx = 0
				dll.momy = 0
			end
		elseif dll.type == MT_EGGSHIELD then P_KillMobj(dll, win.mo, win.mo)
		elseif (dll.type == MT_EGGMOBILE or dll.type == MT_EGGMOBILE or dll.type == MT_EGGMOBILE2
			or dll.type == MT_EGGMOBILE3 or dll.type == MT_EGGMOBILE4 or dll.type == MT_FANG
			or dll.type == MT_CYBRAKDEMON or dll.type == MT_BLACKEGGMAN)
			or (dll.flags & MF_BOSS) then
				if dbg.value then CONS_Printf(win, dll.type + " is a Boss, 2x damage.") end
				P_DamageMobj(dll, win.mo, win.mo, 2)
		else P_DamageMobj(dll, win.mo, win.mo) end
	end
end

// Recreating P_NukeEnemies sparkle spawn.
// Took me a while to get it down the way I wanted.
// Seperated from DXAoE. Somehow went missing during 1.6dev.
local function DXSparkle(win)
	for i = 0, 8, 1 do
		local sparkle = P_SpawnMobjFromMobj(win.mo, 0, 0, 0, MT_SUPERSPARK)
		sparkle.angle = FixedAngle(i*ANGLE_22h)
		P_Thrust(sparkle, sparkle.angle, 40*win.mo.scale)
	end
end

// Cooperating hardware acceleration check! Ain't that just clusters, though?
local function IsInRange(source, target, h, v) // Tatsuru was here, cheers!
    if not (source and source.valid) return end
    if not (target and target.valid) return end
    if not h or not v return end
	if source == target return end
    
    if R_PointToDist2(source.x, source.y, target.x, target.y) > (h * source.radius) return end

    if (target.z < source.z - v * source.height) return end
    if (target.z > source.z + v * source.height) return end
	
	if dbg.value then
		CONS_Printf(target.player, "You were successfully DirectX Booosted by "+source.player.name".")
		CONS_Printf(source.player, "You successfully DirectX Booosted "+target.player.name".")
	end
    return true
end

addHook("PlayerThink", function(player)
	if (player.mo and player.mo.valid) and player.mo.skin != "windows" then
		player.dxready = false
		player.winpopup = 10*TICRATE
		if player.cpanel == nil then player.cpanel = 0 
		elseif player.cpanel != 0 then player.cpanel = 0 end
		return
	elseif (player.mo and player.mo.valid) and player.mo.skin == "windows" then
		if directxboost.value == 1 then
			local plunge = false
			if abs(player.mo.momz) < 0 then plunge = true else plunge = false end
			if player.dxready and player.playerstate == PST_LIVE and player.exiting == 0 then
				if (gametyperules & GTR_STARTCOUNTDOWN and leveltime < CV_FindVar("hidetime").value*TICRATE) or player.powers[pw_nocontrol] != 0 then return end
				if player.cmd.buttons == (BT_CUSTOM2 | BT_JUMP) and player.powers[pw_carry] == 0 then
					soundblaster(player)
					local force = 0
					if P_PlayerInPain(player) then
						// Pain cancel. Full cooldown penalty.
						player.winpopup = $ - 10*TICRATE
						player.mo.state = S_PLAY_JUMP
						P_SetObjectMomZ(player.mo, 15*FRACUNIT)
					elseif player.powers[pw_super] == 0 then
						player.winpopup = $ - 5*TICRATE
						P_SetObjectMomZ(player.mo, 20*FRACUNIT)
						force = 20*FRACUNIT
					elseif player.powers[pw_super] != 0 or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then
						player.winpopup = $ - 2*TICRATE
						if IsSpecialHyper(player) then
							P_SetObjectMomZ(player.mo, 50*FRACUNIT)
							force = 50*FRACUNIT
						else
							P_SetObjectMomZ(player.mo, 35*FRACUNIT)
							force = 35*FRACUNIT
						end
					end
					if (player.charflags & SF_DASHMODE and player.dashmode >= 3*TICRATE) then
						if player.speed/5 <= skins[player.mo.skin].runspeed and player.speed > skins[player.mo.skin].normalspeed then
							P_InstaThrust(player.mo, player.mo.angle, player.runspeed)
						else
							P_InstaThrust(player.mo, player.mo.angle, player.speed/5)
						end
					else
						P_InstaThrust(player.mo, player.mo.angle, player.speed/5)
					end
					if player.powers[pw_super] != 0 or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then
						local poof = P_SpawnGhostMobj(player.mo)
						poof.destscale = player.mo.scale*8
						poof.scalespeed = (player.mo.scale/12)*(player.mo.scale/FRACUNIT)
					else
						for i = 0,3,1
							local r = player.mo.radius/FRACUNIT/4
							local xy = (P_RandomRange(r, -r)*FRACUNIT)
							local z = (P_RandomKey(player.mo.height/FRACUNIT)*FRACUNIT)
							local smokey = P_SpawnMobj(player.mo.x + xy, player.mo.y + xy, player.mo.z + z, MT_SMOKE)
							smokey.scale = player.mo.scale*5
						end
					end
					player.mo.state = S_PLAY_JUMP
					player.pflags = $ | (PF_THOKKED)
					player.pflags = $ & ~PF_SPINNING
					if directplay.value and gametype == GT_COOP and force != 0 then
						for dxdiag in players.iterate
							if IsInRange(player.mo, dxdiag.mo, directplayrangeh.value, directplayrangev.value) and dxdiag.mo != player.mo
							and dxdiag.dxcooptoggle then
								P_SetObjectMomZ(dxdiag.mo, force)
								local dxpoof = P_SpawnGhostMobj(dxdiag.mo)
								dxpoof.destscale = dxdiag.mo.scale*8
								dxpoof.scalespeed = (dxdiag.mo.scale/6)*(dxdiag.mo.scale/FRACUNIT)
							end
						end
					end
					if player.powers[pw_super] != 0 and ((player.hyper and player.hyper.capable) or IsSpecialHyper(player)) then
						DXAoE(player, 1024*FRACUNIT)
						DXSparkle(player)
					end
				elseif player.cmd.buttons == (BT_CUSTOM2 | BT_SPIN) and not P_PlayerInPain(player) and player.powers[pw_carry] == 0 then
					if P_IsObjectOnGround(player.mo) then
						if player.powers[pw_super] == 0 then
							player.winpopup = $ - 5*TICRATE
						elseif player.powers[pw_super] != 0 or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then
							if IsSpecialHyper(player) or (player.hyper and player.hyper.capable) then
								player.winpopup = $ - 4
							else
								player.winpopup = $ - 2*TICRATE
							end
						end
						if IsSpecialHyper(player) and player.powers[pw_super] != 0 then
							if FixedMul(player.speed, player.mo.scale) < 60*FRACUNIT then
								P_InstaThrust(player.mo, player.mo.angle, 240*FRACUNIT)
								S_StartSound(player.mo, sfx_spndsh)
							else
								P_InstaThrust(player.mo, player.mo.angle, player.speed + player.speed/10)
							end
							P_NukeEnemies(player.mo, player.mo, 1024*FRACUNIT)
						else
							if FixedMul(player.speed, player.mo.scale) < 20*FRACUNIT then
								if player.powers[pw_super] != 0 or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then
									P_InstaThrust(player.mo, player.mo.angle, 64*FRACUNIT)
								else
									P_InstaThrust(player.mo, player.mo.angle, 24*FRACUNIT)
								end
								S_StartSound(player.mo, sfx_spndsh)
							elseif FixedMul(player.speed, player.mo.scale) >= 20*FRACUNIT then
								P_InstaThrust(player.mo, player.mo.angle, (4*player.speed)/3)
							end
						end
						S_StartSound(player.mo, sfx_spin)
						player.pflags = $ | (PF_SPINNING)
						player.mo.state = S_PLAY_ROLL
						player.defragment = true
					else
						if player.powers[pw_super] != 0 or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then
							S_StartSound(player.mo, sfx_pinbl1)
							if P_IsObjectInGoop(player.mo) then P_SetObjectMomZ(player.mo, -35*FRACUNIT)
							elseif FixedMul(player.speed, player.mo.scale) >= 55*FRACUNIT then P_SetObjectMomZ(player.mo, player.speed*-1, plunge)
							else P_SetObjectMomZ(player.mo, -55*FRACUNIT, plunge) end
							if directplay.value and gametype == GT_COOP then
								for dxdiag in players.iterate
									if IsInRange(player.mo, dxdiag.mo, directplayrangeh.value, directplayrangev.value) and dxdiag.mo != player.mo and dxdiag.dxcooptoggle then
										if P_IsObjectInGoop(dxdiag.mo) then P_SetObjectMomZ(dxdiag.mo, -35*FRACUNIT)
										elseif FixedMul(dxdiag.speed, dxdiag.mo.scale) >= 55*FRACUNIT then P_SetObjectMomZ(dxdiag.mo, dxdiag.speed*-1)
										else P_SetObjectMomZ(dxdiag.mo, -55*FRACUNIT) end
										local dxpoof = P_SpawnGhostMobj(dxdiag.mo)
										dxpoof.destscale = dxdiag.mo.scale*8
										dxpoof.scalespeed = (dxdiag.mo.scale/6)*(dxdiag.mo.scale/FRACUNIT)
									end
								end
							end
						elseif player.powers[pw_super] == 0 then
							S_StartSound(player.mo, sfx_inurpc)
							local agp = (player.speed*2)/3
							if P_IsObjectInGoop(player.mo) then P_SetObjectMomZ(player.mo, -15*FRACUNIT)
							elseif FixedMul(player.speed, player.mo.scale) >= 55*FRACUNIT then P_SetObjectMomZ(player.mo, agp*-1, plunge)
							else P_SetObjectMomZ(player.mo, -35*FRACUNIT, plunge) end
							if directplay.value and gametype == GT_COOP then
								for dxdiag in players.iterate
									if IsInRange(player.mo, dxdiag.mo, directplayrangeh.value, directplayrangev.value) and dxdiag.mo != player.mo and dxdiag.dxcooptoggle then
										if P_IsObjectInGoop(dxdiag.mo) then P_SetObjectMomZ(dxdiag.mo, -15*FRACUNIT)
										elseif FixedMul(dxdiag.speed, dxdiag.mo.scale) >= 55*FRACUNIT then P_SetObjectMomZ(dxdiag.mo, agp*-1)
										else P_SetObjectMomZ(dxdiag.mo, -35*FRACUNIT) end
										local dxpoof = P_SpawnGhostMobj(dxdiag.mo)
										dxpoof.destscale = dxdiag.mo.scale*8
										dxpoof.scalespeed = (dxdiag.mo.scale/6)*(dxdiag.mo.scale/FRACUNIT)
									end
								end
							end
						end
						if (player.charflags & SF_DASHMODE and player.dashmode >= 3*TICRATE) then
							player.winpopup = $ + 2*TICRATE
						end
						player.winpopup = $ - 3*TICRATE
						player.pflags = $ | (PF_THOKKED)
						player.mo.state = S_PLAY_FALL
					end
					local poof = P_SpawnGhostMobj(player.mo)
					poof.destscale = player.mo.scale*8
					poof.scalespeed = (player.mo.scale/12)*(player.mo.scale/FRACUNIT)
					if player.powers[pw_super] != 0 and ((player.hyper and player.hyper.capable) or IsSpecialHyper(player)) then
						DXAoE(player, 1024*player.mo.scale)
						DXSparkle(player)
					end
				elseif player.cmd.buttons & BT_CUSTOM2 and not P_PlayerInPain(player) and not (player.cmd.forwardmove == 0 and player.cmd.sidemove == 0) then
					if player.powers[pw_super] == 0 then
						player.winpopup = $ - 8*TICRATE
						if (player.dashmode >= 3*TICRATE and (player.charflags & SF_DASHMODE)) then
							player.winpopup = $ + 2*TICRATE
						end
					elseif player.powers[pw_super] != 0 or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then
						if player.powers[pw_super] != 0 and ((player.hyper and player.hyper.capable) or IsSpecialHyper(player)) then
							DXAoE(player, 1024*player.mo.scale)
							DXSparkle(player)
							player.winpopup = $ - 3*TICRATE
						else
							player.winpopup = $ - 5*TICRATE
						end
					end
					if not (player.dashmode >= 3*TICRATE and (player.charflags & SF_DASHMODE)) then
						if player.powers[pw_sneakers] != 0 then
							player.cpanel = $ + 5*TICRATE
						elseif player.powers[pw_super] != 0 or player.powers[pw_invulnerability] != 0 or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then
							player.cpanel = $ + 4*TICRATE
						else
							player.cpanel = $ + 3*TICRATE
						end
					end
					local dir = 0
					if player.powers[pw_carry] == 0 then
						if player.powers[pw_super] != 0 or (player.powers[pw_sneakers] != 0 and mobjscale.value == 1)
						or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then
							if FixedMul(player.speed, player.mo.scale) < 80*FRACUNIT then
								if not (player.cmd.forwardmove == 0 and player.cmd.sidemove == 0) then
									P_InstaThrust(player.mo, player.mo.angle, 80*player.mo.scale)
									dir = $ + 80*FRACUNIT
									if (player.dashmode >= 3*TICRATE and (player.charflags & SF_DASHMODE)) then
										P_Thrust(player.mo, player.mo.angle, 32*player.mo.scale)
										dir = $ + 32*FRACUNIT
									end
								end
							else
								if player.powers[pw_super] != 0 and ((player.hyper and player.hyper.capable) or IsSpecialHyper(player)) then
									P_InstaThrust(player.mo, player.mo.angle, (4*player.speed)/3)
									dir = $ + ((4*player.speed)/3)
								elseif ((player.hyper == nil or not player.hyper.capable) and player.powers[pw_super] != 0) then
									P_InstaThrust(player.mo, player.mo.angle, (11*player.speed)/10)
									dir = $ + ((11*player.speed)/10)
								else
									P_InstaThrust(player.mo, player.mo.angle, player.speed)
									dir = $ + player.speed
									if (player.dashmode >= 3*TICRATE and (player.charflags & SF_DASHMODE)) then
										P_Thrust(player.mo, player.mo.angle, player.speed/5)
										dir = $ + player.speed/5
									end
								end
							end
						else
							if FixedMul(player.speed, player.mo.scale) < 50*FRACUNIT then
								if not (player.cmd.forwardmove == 0 and player.cmd.sidemove == 0) then
									P_InstaThrust(player.mo, player.mo.angle, 50*player.mo.scale)
									dir = $ + 50*FRACUNIT
									if (player.dashmode >= 3*TICRATE and (player.charflags & SF_DASHMODE)) then
										P_Thrust(player.mo, player.mo.angle, player.speed/6)
										dir = $ + player.speed/6
									end
								end
							else
								P_InstaThrust(player.mo, player.mo.angle, player.speed)
								dir = $ + player.speed
								if (player.dashmode >= 3*TICRATE and (player.charflags & SF_DASHMODE)) then
									P_Thrust(player.mo, player.mo.angle, player.speed/6)
									dir = $ + player.speed/6
								end
							end
						end
					end
					if (player.dashmode >= 3*TICRATE and (player.charflags & SF_DASHMODE)) then
						local poof = P_SpawnGhostMobj(player.mo)
						poof.destscale = player.mo.scale*8
						poof.scalespeed = (player.mo.scale/6)*(player.mo.scale/FRACUNIT)
						S_StartSound(player.mo, sfx_musmn2)
					end
					if directplay.value and gametype == GT_COOP then
						for dxdiag in players.iterate
							if IsInRange(player.mo, dxdiag.mo, directplayrangeh.value, directplayrangev.value) and dxdiag.mo != player.mo and dxdiag.dxcooptoggle and dxdiag.speed > 0 then
								P_InstaThrust(dxdiag.mo, dxdiag.mo.angle, dir)
								if dxdiag.mo.eflags & MFE_SPRUNG == 0 then
									P_SetObjectMomZ(dxdiag.mo, 0)
								end
								local dxpoof = P_SpawnGhostMobj(dxdiag.mo)
								dxpoof.destscale = dxdiag.mo.scale*8
								dxpoof.scalespeed = (dxdiag.mo.scale/6)*(dxdiag.mo.scale/FRACUNIT)
							end
						end
					end
					S_StartSound(player.mo, sfx_zoom)
					if player.mo.eflags & MFE_SPRUNG == 0 then
						P_SetObjectMomZ(player.mo, 0)
					end
					if player.powers[pw_super] != 0 or (player.solchar and player.solchar.istransformed and player.solchar.istransformed > 1) then S_StartSound(player.mo, sfx_winxbt)
					elseif player.powers[pw_super] == 0 then S_StartSound(player.mo, sfx_wincbt) end
				end
				player.dxready = false
			end
		end
		// DirectX Boost visual effects.
		if player.cpanel > 0 and player.playerstate == PST_LIVE then
			if leveltime%3 == 1 and not (player.hyper and player.hyper.capable and player.powers[pw_super] != 0) then
				// Thanks, Shine!
				local afterimage = P_SpawnGhostMobj(player.mo)
				afterimage.flags = $ & ~MF_NOCLIPHEIGHT
				afterimage.destscale = (player.mo.scale*16)*(player.mo.scale/FRACUNIT)
				local r = player.mo.radius/FRACUNIT
				local xy = (P_RandomRange(r, -r)*FRACUNIT)
				local z = (P_RandomKey(player.mo.height/FRACUNIT)*FRACUNIT)
				local smokey = P_SpawnMobj(player.mo.x + xy, player.mo.y + xy, player.mo.z + z, MT_SMOKE)
				smokey.colorized = true
				smokey.color = player.mo.color
				local xbox = P_SpawnMobj(player.mo.x + xy, player.mo.y + xy, player.mo.z + z, MT_WATERZAP)
			end
			// Dashmode? Talk about booting Windows in 5 seconds.
			if player.dashmode < 3*TICRATE and player.dashmode > 0 and (player.charflags & SF_DASHMODE) then
				player.dashmode = $ + 1
			end
			player.cpanel = $ - 1
		end
		// Cooldown stuff.
		if player.winpopup != 10*TICRATE and not player.dxready and player.exiting == 0 then
			if player.winpopup < 10*TICRATE then
				if player.rstrui and player.winpopup < (19*TICRATE)/2 then
					player.winpopup = (19*TICRATE)/2
					player.rstrui = false
				end
				// If Hyper, speed it up more!
				if player.powers[pw_super] != 0 and ((player.hyper and player.hyper.capable) or IsSpecialHyper(player)) then
					player.winpopup = $ + 3
				end
				// Else, just go through normally.
				if (player.powers[pw_invulnerability] != 0 or player.powers[pw_sneakers] != 0)
				or player.gotflag == 1 or player.gotflag == 2 then
					player.winpopup = $ + 2
				else
					player.winpopup = $ + 1
				end
				// Dashmode... overclocking... be careful.
				if player.dashmode >= 3*TICRATE and (player.charflags & SF_DASHMODE) then
					player.winpopup = $ + 2
				end
			elseif player.winpopup > 10*TICRATE then player.winpopup = 10*TICRATE
			end
		elseif player.winpopup == 10*TICRATE then player.dxready = true
		end
		if not player.dxready and player.winpopup == 10*TICRATE then
			S_StartSound(player.mo, sfx_winrbt)
			player.dxready = true
		end
	end
end)

// Rings, Coins and Blue Spheres reduce cooldown. Deduplicated.
local function resourcemeter(special, play)
	local com = play.player
	if (com.mo and com.mo.valid) and com.mo.skin != "windows" then return
	elseif (com.mo and com.mo.valid) and com.mo.skin == "windows" then
		if com.winpopup < (10*TICRATE - TICRATE/4) then com.winpopup = $ + TICRATE/4
		elseif com.winpopup > (10*TICRATE - 3*TICRATE/4) and com.winpopup != 10*TICRATE then
			com.winpopup = 10*TICRATE - 1
			if not S_SoundPlaying(com.mo, sfx_winrbt) then
				S_StartSound(com.mo, sfx_winrbt)
			end
		end
	end
end
addHook("TouchSpecial", resourcemeter, MT_RING)
addHook("TouchSpecial", resourcemeter, MT_COIN)
addHook("TouchSpecial", resourcemeter, MT_BLUESPHERE)

// DirectX Boost lasts ~33% longer if you're riding a ball.
addHook("MobjThinker", function(rock)
	if rock and rock.valid and rock.tracer and rock.tracer.valid and rock.tracer.skin == "windows" then	
		local user = rock.tracer.player
		if user.powers[pw_carry] != CR_ROLLOUT then return end
		if user.cpanel > 0 then
			if (user.cmd.forwardmove != 0 or user.cmd.sidemove != 0) then
				P_Thrust(rock, rock.tracer.angle, 5*FRACUNIT)
			end
			if leveltime%3 == 1 then
				user.cpanel = $ + 1
			end
		end
	end
end, MT_ROLLOUTROCK)

// Also spike busting.
local function regedit32(malware, defender)
	if (defender.skin ~= "windows" or not defender.player) then return end
	if malware.z + malware.height < defender.z + defender.momz or malware.z > defender.z + defender.momz + defender.height then return nil end
	local ram = defender.player
	if (ram.powers[pw_super] != 0 or (ram.solchar and ram.solchar.istransformed and ram.solchar.istransformed > 1) or ram.powers[pw_invulnerability] != 0) or ram.cpanel > 0
	or (ram.speed > 16*FRACUNIT and ram.pflags & (PF_SPINNING)) or ram.speed > 48*FRACUNIT then
		P_KillMobj(malware, defender, defender, nil, nil)
	end
end
addHook("MobjCollide", regedit32, MT_SPIKE)
addHook("MobjCollide", regedit32, MT_WALLSPIKE)

// Minecart shenanigans.
addHook("MobjDeath", function(kart)
	if kart and kart.valid and kart.target and kart.target.player
	and kart.target.skin == "windows" and kart.target.health then
		kart.target.player.powers[pw_carry] = 0
		kart.target.momx = kart.momx
		kart.target.momy = kart.momy
		kart.target.momz = kart.momz
		if kart.target.player.cmd.buttons == (BT_SPIN | BT_JUMP) then
			kart.target.state = S_PLAY_PAIN
		else
			kart.target.state = S_PLAY_PAIN
			P_DoPlayerPain(kart.target.player)
		end
		S_StartSound(kart.target, sfx_shldls)
		kart.target = nil
		kart.state = S_NULL
		return true
	end
end, MT_MINECART)

addHook("MobjThinker", function(kart)
	if kart and kart.valid and kart.target and kart.target.valid
	and kart.target.player and kart.target.type == MT_PLAYER
	and kart.target.skin == "windows" and kart.target.health then
		if kart.target.player.cmd.buttons == (BT_SPIN | BT_JUMP) then
			P_KillMobj(kart)
		end
	end
end, MT_MINECART)

// You can always tell a Texan... but can't tell the saloon door much.
// Oh, and Kart-esque monitor crushing.
addHook("MobjMoveCollide", function(win, thingy)
	if not (win or win.player) return end
	if not (thingy and thingy.valid) return end
	if win.skin == "windows" and not (win.z > thingy.z+thingy.height) and not (thingy.z > win.z+win.height)
		if ((thingy.type == MT_SALOONDOOR) or (thingy.type == MT_SALOONDOORCENTER))
			if thingy.valid and thingy.health and not (thingy.flags & MF_NOCLIP)
			and ((win.player.powers[pw_super] != 0 or win.player.powers[pw_invulnerability] != 0 or (win.player.solchar and win.player.solchar.istransformed and win.player.solchar.istransformed > 1))
			or win.player.cpanel > 0 or FixedMul(win.player.speed, win.scale) > 48*FRACUNIT or (win.player.pflags & (PF_SPINNING) and FixedMul(win.player.speed, win.scale) >= 16*FRACUNIT)) then
				for i = 0,7,1
					local r = thingy.radius/FRACUNIT
					local xy = (P_RandomRange(r, -r)*FRACUNIT)
					local z = (P_RandomKey(thingy.height/FRACUNIT)*FRACUNIT)
					local wood = P_SpawnMobj(thingy.x + xy, thingy.y + xy, thingy.z + z, MT_WOODDEBRIS)
					wood.momx = win.momx/3
					wood.momy = win.momy/3
					wood.scale = thingy.scale
				end
				if (thingy.type == MT_SALOONDOOR) then
					thingy.flags = $|MF_NOCLIP
					P_KillMobj(thingy, h, h)
				elseif (thingy.type == MT_SALOONDOORCENTER) then
					thingy.flags = $|MF_NOCLIP
				end
				S_StartSound(win, sfx_wbreak)
				return false
			end
		elseif (thingy.flags & MF_MONITOR) and win.player.powers[pw_sneakers] != 0 and mobjscale.value and P_IsObjectOnGround(win) and win.scale > thingy.scale then
			P_DamageMobj(thingy, win, win)
			return false
		end
	end
end, MT_PLAYER)

// Busting through FoFs.
// Thanks for the blessing, SMS Alfredo!
local function bustcheck(fof)
	return fof and fof.valid and fof.flags & (FF_EXISTS|FF_BUSTUP) == (FF_EXISTS|FF_BUSTUP)
end

addHook("PlayerThink", function(player)
	if (player.mo and player.mo.valid) and player.mo.skin == "windows" then
		local sectorthing = P_PlayerTouchingSectorSpecial(player, 0, 0)
		if sectorthing and sectorthing.valid then
			for fof in sectorthing.ffloors()
				if fof.valid and (abs(fof.topheight - player.mo.z) <= 40*player.mo.scale
				or abs(fof.bottomheight - player.mo.z+player.mo.height) <= 40*player.mo.scale)
				and fof.master and fof.master.valid and fof.master.flags & ML_NOCLIMB
				and (fof.master.special == 153 or fof.master.special == 180
				or (fof.master.special >= 150 and fof.master.special <= 152)
				or (fof.master.special >= 190 and fof.master.special <= 195)
				or fof.master.special == 176 or fof.master.special == 177) then
					//if P_IsObjectOnGround(player.mo) then -- this appears to be broken.
						player.pflags = $ | (PF_STARTDASH | PF_SPINNING)
						P_InstaThrust(player.mo,0,0)
					//end
				end
			end
		end
		// The Guy, The Lua was here too.
		if bustcheck(player.mo.floorrover) and player.mo.z == player.mo.floorrover.topheight
			EV_CrumbleChain(player.mo.floorrover)
		end
		if bustcheck(player.mo.ceilingrover) and (player.mo.z + player.mo.height) == player.mo.ceilingrover.bottomheight
			EV_CrumbleChain(player.mo.ceilingrover)
		end
	end
end)